  #define _16F84		;define target microcontroller

;  #define _12C509

; Last update 4/22/02  Jim Remington jim@uoxray.uoregon.edu

;**********************************************************************

; PIC Solar Water Heater Monitor

;**********************************************************************

;

; PIC 12c509A/16F84A for Dallas/Maxim DS1822 temperature sensor interface 

; Intended for continuous (1 per second) readings of input and output

; temperatures of a solar water heating panel. Also records switch closures

; of a water flow meter (Omega Engineering) on the input side, up to about

; 15 per second.

;

; Author: Jim Remington jim@uoxray.uoregon.edu

; DS1822 code modified from DS1820 code by Steve Marchant

;

; The current configuration reads 2 temp sensors and 1 switch (water flow meter)

; Output is RS232 (inverted TTL) strings with current sensor readings

; at 9600 baud to a PC laptop monitor.

;

; RS232 ASCII string is: 

;	CRLF sent first, to prevent premature line erasure on LCD display

;	P,xNNN.N  (leading blank supressed). 

; where

;	x=(-), P=Port bit # of attached sensor

;	or,

;	9,SS where SS = # of debounced switch closures in last second.

; 	One measurement per second (roughly)

;

; current configuration: RS232 output on  PORTB.2 

;	indicating device # and temperature (C), 

;	Temp range -55.0 to +125.0, resolution 0.5 deg.

;

; DS1822 #1 I/O on PORTB.0 (need 2-4.7K pullup)

; DS1822 #2 I/O on PORTB.1 (ditto)

; DS1822  on PORTB.4 and 5 are possible but are not implemented 

;         (currently, test for max sensor # is RS232 port pin #)

; 

; for 12c509, don't forget that GP3 is input only.

;

; DS1822 devices NOT PRESENT are automatically skipped.

;

; in line-code for 12c5xx compatibility, CALL DEPTH limited to 2

;

; tested using both 12c509A (internal 4 MHz RC osc) and 16f84A (4 MHz crystal)

;					

;**********************************************************************

;

	ifdef	_16F84

	list      p=16f84A            ; list directive to define processor

	#include <p16f84A.inc>        ; processor specific variable definitions



	__CONFIG   _CP_OFF & _WDT_OFF & _XT_OSC & _PWRTE_ON

 

OSCCAL	equ	5		;dummy OSCCAL for 16f84

	endif

;

	ifdef	_12C509

	list	p=12C509A

	#include <p12c509A.inc>        ; processor specific variable definitions



	__CONFIG   _CP_OFF & _WDT_OFF & _IntRC_OSC & _MCLRE_OFF



	endif



; PORT and BIT definitions

;



PORTB		EQU	6	;GPIO

switch		equ	3	;water flow meter switch on GPIO.3

SEROUT		equ	2	;GPIO.2 RS232 out (Inverted TRUE) 

;				;2 temp sensors on GPIO.0,GPIO.1



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	RAM DEFINITIONS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;

	cblock	0x0C	;could move this down on the 12c509...

;

; DS1822 variables

; common to all sensors

;

DS_POINT	;bit number of current DS1822 DQ

DS_BIT		;1 at bit position for current DS1822 I/O pin

TrisB		;copy of tristate register (which is read only)

;

IN_BYTE		;byte result from DS1822

TMP0		;PAUSE counter

TMP1		;temp W, DLY250 count, 8 bit count, temp sign

TMP2		;temp fraction bits, RX, TX data temp

;

DSTempL		;result low byte

DSTempH		;result high byte

;

; general variables

;

loop_count	;count ~ seconds (updated but not used)

minutes		;16 bit minute counter (updated but not used)

m256		; high byte (about 45 days, updated but not used)

sw		;GPIO.3 status

sw_cnt		; switch closures since last check

;

t1	;temps used by 

t2	;ascii format/output 

t3

;

RS1	;RS232 temp char

RS2	;RS232 bit count

RS3	;RS232 bit delay

	endc



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	MACRO DEFS (modified from Steve Marchant's) ;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



;	1-WIRE OUTPUT LO

DQLOW:	MACRO

	COMF	DS_BIT,w	; get a 0 at DS I/O pin position

	ANDWF	PORTB,f		; set that pin to 0 on PORTB

	ANDWF	TrisB,f		; set to that bit to output in TRIS register copy

	MOVFW	TrisB

	TRIS	PORTB		;  set pin to O/P

	ENDM



;	1-WIRE INPUT, high impedance bus

DQIN:	MACRO

	MOVFW	DS_BIT		;1 at DS I/O pin position

	IORWF	TrisB,f		; save a copy

	MOVFW	TrisB

	TRIS	PORTB		; DQ BIT NOW I/P

	ENDM



PAUSE:	MACRO	DLYF		; delay for MULTIPLE OF 5 microseconds (1275 max)

	MOVLW	(DLYF / 5) - 1	;  (.10 gives 6 cycles)

	MOVWF	TMP0

	local	Loop

Loop	NOP

	NOP

	DECFSZ	TMP0,f

	GOTO	Loop

	ENDM



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	STARTUP CODE		;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	ORG 	0

	movwf	OSCCAL		;save clock calibration

;

; I/O Port

;

	clrf	DS_POINT	;set pointer to current DS sensor

	movlw	1		; assume on PORTB.0

	movwf	DS_BIT		; save bit indicator

	MOVLW	(1<<switch) + 1		; bits 0 and (3) input

	movwf	PORTB

	movwf	TrisB		;save copy of port control reg

	TRIS	PORTB

;

	ifdef	_12C509

	movlw	B'10000110'	;bit7=1: no wakeup; bit 6=0 weak pullups

				;bit5=0 T0CS=CLK

				;bit3=0 PSA->TMR0

				;bit2-0: prescaler rate 1:128  

	endif



	ifdef	_16F84

	movlw	B'00000110'	;bit7=0 weak pullups ON

				;bit5=0 T0CS=CLK

				;bit3=0 PSA->TMR0

				;bit2-0: prescaler rate 6=1:128

	endif

 

	OPTION

;	

	clrf	loop_count

	clrf	minutes

	clrf	m256

;

	call	dly1s		;wait for display, etc. to stabilize

	clrf	sw		;clear switch flags and counter

	clrf	sw_cnt

;



	movlw	'V'		;announce with version number

	call	rsout

	movlw	'3'

	call	rsout

	movlw	'.'

	call	rsout

	movlw	'2'

	call	rsout		; no CR/LF yet...

	goto	gettemp



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	SUBROUTINES

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



;

; delay 1 second. In the meantime check PORTB.switch (3) for switch closure

;



dly1s	movlw	.61		;delay 61x16.384 ms = 0.999424s plus change

	movwf	TMP0

	clrf	TMR0		;initialize timer

;

;	check PORTB.switch, store switch state

;

dly1	btfss	PORTB,switch	; hi=open	

	bsf	sw,switch	; or, record switch closed (sw.3=1)



; DELAY 16 ms



dly16ms	btfss	TMR0,7		;check for timer overflow to bit 7

	goto	dly16ms		; (prescaler=128) so 16384 cycles

	clrf	TMR0		;and start from scratch

;

; check switch (3) again after 16 ms to debounce

; timer is still incrementing during this code

;

	btfss	sw,switch	; closed last pass?

	goto	dly2		; no, do nothing this pass

	btfsc	PORTB,switch	; is it still closed? count it

	goto	dly2		; open, clear flags and continue countdown

	btfss	sw,7		; check whether we counted this, sw.7=1

	incf	sw_cnt,f	;  bump count and 

	bsf	sw,7		;   set flag for having counted

	goto	dly2+1		; skip next if counted

dly2	clrf	sw		;  clear debounce switch/count flag



; loop

	decfsz	TMP0,f

	goto 	dly1

	retlw 0



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	DS1822 support routines	;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; modified from DS1820 code on David Tait's PIC page

; http://www.thepicarchive.cwc.net/dtpa/

; by: Steve.Marchant@Nottingham.ac.uk

;     Thanks, Steve!

;

; Steve's comments, modified:

; --------------------------------------------------------

;	DS1822 support for PIC16F84

;	DS1822 DQ pin connected to PORTB on pin DSBIT

;

;	Requires TMP0,1 & 2 temporary working registers.

;	Results are stored in DSTempH and DSTempL registers.

;

;	Call DSCHECK every time round in your

;	main loop and test the W reg to see if

;	new data available (W=0). 

;	(AFTER issuing ConvertT command)

;

;	NOTE parts of this code are timing critical

;	and must be run with INTERRUPTS turned OFF!

;

;	RETURNS W<>0 if data not ready

;	        W=0 if a new measurement has been made

;	Output in DSTempL & DSTempH

;	TMP0, TMP1, TMP2 used

;

DSCHECK:

        CALL    DSRX8		; CHECK DS STATUS

	MOVLW	1

	ADDWF	IN_BYTE,w	; 255=READY => 0=READY

        BTFSS   STATUS,Z	; Z set = READY, get temp

        retlw 1			; W non zero = not ready

;

;	Get current temperature

;

DSTEMP:	

	CALL	DSRESET

	MOVLW	H'CC'		; SKIP PROM

	CALL	DSTX8

	MOVLW	H'BE'		; READ SCRATCH PAD

	CALL	DSTX8

	CALL	DSRX8		; get temp low byte

	MOVFW	IN_BYTE

	MOVWF	DSTempL		; Store it

	CALL	DSRX8		; get temp high byte

	MOVFW	IN_BYTE

	MOVWF	DSTempH		; Store it

	retlw 0

;

;	Start the next conversion

;

CONVTT:	CALL	DSRESET		; SEEN ENOUGH, RESTART

	MOVLW	H'CC'		; SKIP PROM

	CALL	DSTX8

	MOVLW	H'44'		; START CONVERT

	CALL	DSTX8

	retlw 0

; ------------------------------------------------------------

;	DS1820 subroutines 

; 	DSRESET destroys TMP0,1

;	returns 0 in IN_BYTE if valid response

;

DSRESET:

	DQLOW

	PAUSE	.600		;minimum 480 usec

	DQIN

	PAUSE	.70		; wait 70us for presence bit (min 60)

	MOVF	PORTB,W

	ANDWF	DS_BIT,W

	MOVWF	IN_BYTE		;save response (0=device present)

	PAUSE	.500		;minimum 480 after reset pulse

	retlw 0



DSTX8:	MOVWF	TMP2		; save data to transmit

	MOVLW	.8

	MOVWF	TMP1		; LOOP COUNTER



DSTXLP:	DQLOW

	PAUSE	.10		; bus release within 15 usec

	RRF	TMP2,f

	MOVFW	PORTB

	IORWF	DS_BIT,w	; get correct bit to O/P

	BTFSC	STATUS,C

	MOVWF	PORTB		; DQ HI IF BIT WAS 1

				;otherwise keep it low for

	PAUSE	.70		;minimum of 60 usec for write slot

	DQIN

	NOP

	DECFSZ	TMP1,f

	GOTO	DSTXLP

	retlw 0

;

DSRX8:	MOVLW	.8

	MOVWF	TMP1



DSRXLP: CLRC

	RRF	IN_BYTE,f	

	DQLOW

	NOP			;minimum of 1 usec

	DQIN			; then hi-Z

	PAUSE	.10	

	MOVF	PORTB,W		; read toward end of 15 usec slot (recommended)

	ANDWF	DS_BIT,W	; ->0 IF DATA BIT LOW

	BTFSS	STATUS,Z

	BSF	IN_BYTE,7	;set high bit of IN_BYTE if W<>0

	PAUSE	.60

	DECFSZ	TMP1,f

	GOTO	DSRXLP

	retlw 0

;

; RS232 output byte in W, PORTB.SEROUT - 9600 baud (INVERTED TRUE)

; uses ram locations RS1,RS2,RS3

;

rsout:	movwf	RS1	;temp char out

	movlw	.8

	movwf	RS2	;bit counter

	bsf	PORTB,SEROUT

	call	rs_wait

rsl:	btfss 	RS1,0	;1(2) 5 cycles either path

	goto $+3	;2(0)

	bcf	PORTB,SEROUT	;0(1)

	goto $+4	;0(2)

	bsf	PORTB,SEROUT	;1(0)

	nop		;1(0)

	nop		;1(0)

	rrf	RS1,f	;1

	call	rs_wait	;2

	decfsz	RS2,f	;1

	goto 	rsl	;2	total 104 cycles/bit

	bcf	PORTB,SEROUT

;

; 	9600 baud wait @ 4Mhz = 104 cycles/bit. burn 93 here

;

rs_wait: 

	movlw	.30 	;1

	movwf	RS3	;1

	decfsz	RS3,f	;2+29*3=89

	goto	$-1

	retlw 0		;2



;888888888888888888888888

;8	MAIN LOOP	8

;888888888888888888888888



gettemp:

;

; check whether switch on PORTB.3 has been pressed 

;

	movfw	sw_cnt		; were counts accumulated in last delay?

	BZ	gettm1		; no, skip output

;

; output line to RS232 9,SS	;where s=#switch closures in delay loop

;

; terminate previous output line (in case of LCD display)

;

	movlw	.13		;CR/LF

	call	rsout

	movlw	.10

	call	rsout

;

	movlw	'9'

	call	rsout

	movlw	','

	call	rsout



	movfw	sw_cnt

	clrf	sw_cnt		;clear switch counter

;

; format W as 2 ascii digits NN (60 is max possible for switch test in delay loop)

;

	movwf	t3

	movlw   .10

	clrf    t1	;clear digit counter

NN10:	incf    t1,f	

	subwf   t3,f	; subtract .10

	btfsc   STATUS,C	;C=0, result is negative

	goto	NN10

	addwf   t3,f	; add it back

	movlw   '0'-1	; make ascii (minus extra decrement)

	addwf   t1,w

	call	rsout

	movlw   '0'

        addwf   t3,w

        call 	rsout	; output units

;

; loop over sensors...

;

gettm1	call	DSRESET		; check presence of sensor

	movf	IN_BYTE,f	; 0=sensor active

	BNZ	next_DS		; try next one

	call	CONVTT		;start it up

	call	dly1s		; DELAY overlaps conversion time

gettm2	call 	DSCHECK		; Check temp conversion done

	iorlw	0		; if W<>0 then not ready

	BNZ	gettm2

;

	movfw	DSTempH

	movwf	TMP0		;save sign

	btfss	TMP0,7		; if negative, two's complement

	goto	reduceP

;

	comf	DSTempL,f	; 16 bit complement and add 1

	comf	DSTempH,f

	incf	DSTempL,f

	btfsc	STATUS,Z

	incf	DSTempH,f

;

; 11 bits, convert to 8

;

reduceP	rrf	DSTempH,f

	rrf	DSTempL,f

	rrf	DSTempH,f

	rrf	DSTempL,f

	rrf	DSTempH,f

	rrf	DSTempL,f ; 8 bits, 0.5 degrees C resolution

;

; terminate previous output line (for LCD display)

;

	movlw	.13		;CR/LF to terminate last line

	call	rsout

	movlw	.10

	call	rsout

;

; format and transmit 

; output pointer #, +/-temp

;

	movlw	'0'		;output sensor #

	addwf	DS_POINT,w

	call	rsout

	movlw	','

	call	rsout		; comma

	movlw	'-'

	btfsc	TMP0,7		;check sign bit

	call	rsout		;output minus sign if set



;

; output temperature in format NNN.N

;

	clrc

	rrf	DSTempL,w	; reduce temp to 7 bits 0-125

	rrf	TMP2,f		; save fraction bit

;

;---------------------------------------------------------

;

; Output Decimal (value in W) with leading zero supression

; uses t1, t2, t3

;

ascout:	clrf    t2 	;clear leading zero flag

	movwf	t3	;save number

        movlw   .100

	clrf    t1	;clear digit counter

asc1:	incf    t1,f	

	subwf   t3,f	; subtract .100 

	btfsc   STATUS,C	;C=0, result is negative

	goto	asc1

	addwf   t3,f	; add it back

	decf    t1,f	; make up for extra inc

	movf    t1,w

	xorwf   t2,w

	btfsc   STATUS,Z ;check for leading zero

	goto    asc2	; skip print if true

	decf    t2,f	; set flag for non-zero digit

	movlw   '0'	; MAKE ASCII

	addwf   t1,w

	call	rsout

asc2:	movlw   .10

	clrf    t1	;clear digit counter

asc3:	incf    t1,f	

	subwf   t3,f	; subtract .10

	btfsc   STATUS,C	;C=0, result is negative

	goto	asc3

	addwf   t3,f	; add it back

	decf    t1,f	; make up for extra inc

	movf    t1,w

	xorwf   t2,w

	btfsc   STATUS,Z ; check for leading zero

	goto    asc4	;skip print if true

	decf    t2,f	; set flag for non-zero digit

	movlw   '0'

	addwf   t1,w

	call	rsout

asc4:	movlw   '0'

        addwf   t3,w

        call 	rsout	; output units

;

	movlw	'.'

	call	rsout

	movlw	'0'		;assume fraction=0

	btfsc	TMP2,7		; check fraction bit of T

	movlw	'5'		;  and output it

	call	rsout

;

; update time, count minutes and overflow (45 days). NOT USED at present

;

	incf	loop_count,f

	movlw	.60

	subwf	loop_count,w	 

	BNZ	next_DS		; minute not up yet

	clrf	loop_count

	incf	minutes,f	; count minutes

	btfsc	STATUS,Z	;  count overflow

	incf	m256,f

;

; point to next available DS sensor

;

next_DS:

	clrc

	rlf	DS_BIT,f	; point to next I/O pin

	incf	DS_POINT,f	;bump pointer

	movlw	SEROUT		; max allowed+1

	subwf	DS_POINT,w

	BNZ	gettemp		;go back for more

	clrf	DS_POINT

	movlw	1

	movwf	DS_BIT

	goto	gettemp

;



	end

